#!/usr/bin/env python
import zmq
import array

import sys
this = sys.modules[__name__]  
this.socket = None

def zezfio_initialise(address):
    context = zmq.Context()
    this.socket = context.socket(zmq.REQ)
    this.socket.connect(address)

def bytes2int(bytes):
    return array.array("i",bytes).pop()

def zezfio_has(msg):
    s = this.socket

    s.send("has.%s"%msg)
    return bytes2int(s.recv())


def zezfio_get(msg):
    s = this.socket

    s.send("get.%s"%msg)
    
    zerrno = bytes2int(s.recv())
   
    if (zerrno >= 0):
        _ = s.recv()
        return s.recv()

def zezfio_set(msg, bytes):
    s = this.socket

    s.send("set.%s"%msg,zmq.SNDMORE)
    s.send(bytes)
    
    return bytes2int(s.recv())


def zezfio_get_ascii(msg):
    s = this.socket

    s.send("get_ascii.%s"%msg)
    
    zerrno = bytes2int(s.recv())
   
    if (zerrno >= 0):
        return s.recv()

def zezfio_set_ascii(msg,string):
    s = this.socket

    s.send("set_ascii.%s"%msg,zmq.SNDMORE)
    s.send(string)    
    return bytes2int(s.recv())



import irpy

from zezfio.io import zio,zscalar,zarray
from zezfio import babel
from zezfio.io.__init__ import build_path

d_instance = dict()
from ctypes import c_int
d_instance["zezfio_id"] = c_int(0)

def update_zezfio_id():
    d_instance["zezfio_id"] = c_int(d_instance["zezfio_id"].value + 1)

{% for category in json_config -%}

class {{ category|capitalize }}(object):

    {% for variable in json_config[category] -%}

    @irpy.lazy_property_mutable
    def {{ variable.name }}_shape(self):
        {% if variable.shape is defined %}
        return {{variable.shape}}
        {% else %}
        return [-1]
        {% endif %}

    @irpy.lazy_property
    def {{ variable.name }}_len(self):
        return babel.shape2len(self.{{ variable.name }}_shape)

    @irpy.lazy_property
    def {{ variable.name }}_path(self):
        return build_path('{{db_path}}','{{ category }}','{{ variable.name }}',
                          len_=self.{{ variable.name }}_len)

    @irpy.lazy_property_mutable
    def {{ variable.name }}_interface(self):
    
        {% if variable.default is defined %}

        data_py = {{ variable.default }}
        return babel.py2interface('{{ variable.type }}',data_py)

        {% else %}
        #IRPY Waring. Nele need to be in a separate variable cause zio can call a C function
        #who will break the genalogy machinery of IRPY
        nele = self.{{ variable.name }}_len
        return zio.db2interface(self.{{variable.name}}_path,'{{ variable.type }}', 
                                len_=nele)

        {% endif %}

    @irpy.lazy_property
    def {{ variable.name }}_len_interface(self):

        {% if variable.type != 'raw' %}

        return babel.len2bytes('{{ variable.type }}',
                               self.{{ variable.name }}_len)
        {% else %}
        return c_int(len(self.{{ variable.name }}_interface))
        {% endif %}

    #This value needs to depend of _interface            #
    #because it can be used to define the size of arrays #
    @irpy.lazy_property
    def {{ variable.name }}(self):
        return babel.interface2py(self.{{ variable.name }}_interface, 
                                  len_=self.{{ variable.name }}_len)

    def set_{{ variable.name }}(self,bytes):

        {% if variable.type != 'raw' %}
        sbytes = len(bytes)
        theobytes = self.{{ variable.name }}_len_interface.value

        if  sbytes != theobytes:
            raise IndexError("Conflic between the spec bytes of the array %i and the given one %i" % (sbytes, theobytes))       

        {% endif %}

        data_interface = babel.bytes2interface('{{ variable.type }}', bytes)
        self.{{ variable.name }}_interface = data_interface

        zio.interface2db(self.{{ variable.name }}_path,'{{ variable.type }}',
                             self.{{ variable.name }}_shape,data_interface)

        update_zezfio_id()

    @irpy.lazy_property
    def {{ variable.name }}_ascii(self):
        return babel.py2ascii(self.{{ variable.name }})

    @irpy.lazy_property
    def {{ variable.name }}_len_ascii(self):
        return c_int(len(self.{{ variable.name }}_ascii))

    def set_ascii_{{ variable.name }}(self,data_ascii):

        sbytes = len(data_ascii)
        theobytes = self.{{ variable.name }}_len_ascii.value

        if  sbytes != theobytes:
            raise IndexError("Conflic between the spec bytes of the array %i and the given one %i" % (sbytes, theobytes))       

        data_interface = babel.ascii2interface('{{ variable.type }}', data_ascii)
        self.{{ variable.name }}_interface = data_interface

        zio.interface2db(self.{{ variable.name }}_path,'{{ variable.type }}',
                             self.{{ variable.name }}_shape,data_interface)

        update_zezfio_id()


    {% endfor %}

#For allowing referance in shape
{{ category }} = {{ category|capitalize }}()
#Quicker than getattr
d_instance[ "{{ category }}" ] = {{ category }}

{%  endfor %}



from zezfio import babel
import array
import os

def db2py(path,t_interface):
    "Return array c_type"

    with open(path, 'r') as f:
        data = f.read().strip()

    #Raw -> Python
    try:
        f = babel.c2stuff[t_interface].str2py
    except KeyError:
        raise TypeError, "Error: cannot convert str to %s" % t_interface
    
    #We need to do the padding
    if not babel.is_char(t_interface):
        py_data = f(data)
    else:
        py_data = f(data,t_interface)

    return py_data

def db2interface(path,t_interface):
    data = db2py(path,t_interface)
    return babel.py2interface(t_interface,data)

#Need to be and array or a string
def interface2db(path, py_scalar):

    if type(py_scalar) == str:
        data = py_scalar
    else:
        data = str(py_scalar[0])

    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))

    with open(path, 'w') as f:
        f.write("%s\n" % data)




import os
dir_path = os.path.dirname(os.path.realpath(__file__))
dll_path = os.path.join(dir_path,"gz2ar.so")

from ctypes import CDLL, get_errno
#Absolute path
dll = CDLL(dll_path,use_errno=True)


from zezfio import babel
from ctypes import c_char

d_erno = {100: "gzopen failed for {file}",
          101: "Your buffer size is to small for the uncompressed data for file {file}",
          102: "There are more values to read than (>{length} asked)",
          103: "I have read less values than requested (<{length})",
          104: "Only 1 or 0 are convertible to bool"}

def check_errno():
    errno = get_errno()
    try:
        message_template = d_erno[errno]
    except KeyError:
        return True
    else:
        message = message_template.format(**locals())
        raise BytesWarning("Error: %s"%message)

def gzip2buffer(file,bytes,header_bytes=200):

    total_bytes = header_bytes + bytes
    buffer_type = (c_char * total_bytes)
    buffer = buffer_type()

    dll.gzip2buffer(file,total_bytes,buffer)
    
    if check_errno():
        return buffer

#This function is IMPURE !!! Buffer will be modified !
def buffer2stuff_impure(t_interface,buffer,length):

    if not babel.is_char(t_interface):

        #Get the C function
        try:
            c_function = getattr(dll,"buffer2%s_impure"%t_interface)
        except AttributeError:
            raise 
            raise BytesWarning("Error: No C function to convert your buffer in this format: %s"%t_interface)
        #Malloc the c_array
        try:
            c_type = babel.c2stuff[t_interface].c_type
        except:
            raise BytesWarning("Error: No C_type related to: %s"%t_interface)
        else:
            c_array_type = ( c_type * length)
            c_array = c_array_type()
         
        #Call the function
        c_function(buffer,length,c_array)

    else:
        #Get the C function
        c_function = getattr(dll,"buffer2char")
    
        #Malloc the c_array
        from ctypes import c_char
        char_size = babel.c2stuff[t_interface].c_size

        real_length = length*char_size

        c_array_type = ( c_char * real_length)
        c_array = c_array_type()
    
        #Call the function
        from ctypes import c_size_t
        c_function(buffer, length,c_size_t(char_size), c_array)

    if check_errno():
        return c_array


import gzip

def db2interface(path,t_interface,nele):
    with gzip.open(path,"r") as f:
        f.readline()
        shape = map(int,f.readline().split())
        dlen = babel.shape2len(shape)

    if dlen != nele:
        raise BytesWarning("Error: Asked number of elements (%s) differ from the file header (%s)"%(dlen,nele))

    try:
        str_length = nele * babel.c2stuff[t_interface].str_size
    except KeyError:
        raise TypeError("Error: No size for %s" % t_interface)
    else:
        buffer = gzip2buffer(path,str_length)

    return buffer2stuff_impure(t_interface,buffer,nele) 


#Need to be and array or a string
def interface2db(path, shape, t_interface, data_interface):

    header = [str(len(shape)), 
              "  ".join(map(str,shape))]

    if babel.is_char(t_interface):
        lstr = "".join(data_interface).split('\x00')
    else:
        lstr = map(str,data_interface.tolist())        

    with gzip.open(path, 'w') as f:
        f.write("%s\n" % "\n".join(header+lstr))



import irpy
import json
import os
def importCode(code,name):

    import sys,imp

    module = imp.new_module(name)
    compiled_code = compile(code, '<string>', 'exec')
    exec compiled_code in module.__dict__
    sys.modules["module"] = module

    return module

def check_list(l,mode="folder"):

    if mode == "folder":
        fct = os.path.isdir
    elif mode == "file":
        fct = os.path.isfile

    l_error = [i for i in l if not fct(i)]
    if l_error:
        raise KeyError("{0} need to be a {1}".format(l_error,mode))


class Generator(object):

    @irpy.lazy_property_leaves(mutables=["db_path","list_folder_user","list_file_user"])
    def __init__(self,arguments):

        db_path_rel = arguments["<ezfio_folder>"]
        self.db_path = os.path.abspath(db_path_rel)

        try:
            l = arguments["--dir"]
            check_list(l,mode="folder")
        except KeyError:
            self.list_folder_user = []
        else: 
            self.list_folder_user = l

        try:
            l = arguments["--file"]
            check_list(l,mode="file")
        except KeyError:
            self.list_file_user = []
        else: 
            self.list_file_user = l

    @irpy.lazy_property
    def list_file(self):
        
        l = self.list_file_user
        for folder in self.list_folder_user:
            l += [os.path.join(folder,i) for i in os.listdir(folder)]
        
        if not l:
            raise KeyError("Provide ether list_file_user or list_folder_user")
        else:
            check_list(l,mode="file")
        return l

    @irpy.lazy_property
    def json_config(self):
        data = dict()

        for path in self.list_file:
            with open(path) as f:
                data.update(json.load(f))

        return data

    @irpy.lazy_property
    def jinja_env(self):
        file_dir = os.path.dirname(__file__)
        template_dir_rel = os.path.join(file_dir,"template")
        template_dir = os.path.abspath(template_dir_rel)
        
        from jinja2 import Environment, FileSystemLoader
        return Environment(loader=FileSystemLoader(template_dir))

    @irpy.lazy_property
    def template_server(self):
        if not self.db_path:
            raise KeyError("You need to set the db_path to generate the server")

        template = self.jinja_env.get_template('management.jinja2.py').render(json_config=self.json_config,
                                                                       db_path=self.db_path)
        return template

    @irpy.lazy_property
    def template_fortran(self):

        from zezfio.babel import c2stuff
        template = self.jinja_env.get_template('management.jinja2.py').render(json_config=self.json_config,
                                                                       c2stuff=c2stuff)
        return template

    @irpy.lazy_property
    def d_instance(self):
        m = importCode(self.template_server, "zezfio_server")
        return m.d_instance








import os

def build_path(root_path,category,raw_name,len_=-1):
    if len_ == -1:
        name = raw_name
    else:
        name = "%s.gz"%raw_name

    path = os.path.join(root_path, category.lower(), name.lower())
    return path


import zscalar
import zarray
from zezfio import babel
import os

def db2interface(path,t_interface,len_):

    if babel.is_raw(t_interface):
        with open(path, 'r') as f:
            data = f.read()
        return data
    elif len_ == -1:
        return zscalar.db2interface(path,t_interface)
    else:
        return zarray.db2interface(path, t_interface, len_)


#Data_interface is a string or and array
def interface2db(path,t_interface,shape,data_interface):

    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))

    if babel.is_raw(t_interface):
        with open(path, 'w') as f:
            f.write("%s\n" % data_interface)
    elif shape == [-1]:
        zscalar.interface2db(path,data_interface)
    else:
        zarray.interface2db(path,shape, t_interface, data_interface)






from ctypes import c_int,c_long,c_float,c_double,create_string_buffer
from collections import namedtuple

def strbool2int(b):
  "1 is True, 0 is False"
  if b in 'Tt1':
     return 1
  elif b in 'Ff0':
     return 0
  else:
     raise TypeError("strbool2int : %s should be in (T|F)"%b)

def ljust2(str_,t_interface):
  padding = int(t_interface[5:-1])
  return str_.ljust(padding)

#c_type is for allocate the array for the c binding
Convert = namedtuple("Convert", ["str_size", "c_size", "c_type","py_array_code", "fortran_type", "str2py"])
c2stuff = {
    "bool"   :   Convert(24,  4,  c_int,    'i', "LOGICAL",          strbool2int ),
    "int"    :   Convert(24,  4,  c_int,    'i', "INTEGER",          int         ),
    "long"   :   Convert(24,  8,  c_long,   'l', "INTEGER*8",        int         ),
    "float"  :   Convert(24,  4,  c_float,  'f', "REAL",             float       ),
    "double" :   Convert(32,  8,  c_double, 'd', "DOUBLE PRECISION", float       )
}

for i in range(3000):
  c2stuff["char[%d]"%i] = Convert(i+1,i,create_string_buffer,'c', "CHARACTER*(%d)"%i, ljust2)

def is_raw(t_interface):
  return "raw" in t_interface

def is_char(t_interface):
  return "char[" in t_interface

def is_string(t_interface):
  return any([is_char(t_interface),is_raw(t_interface)])

from operator import mul
def shape2len(l):
  return reduce(mul, l)

def len2bytes(t_interface,len_=-1):

  if is_char(t_interface):
    return c_int(int(t_interface[5:-1]))
  else:
    nele = len_ if len_ != -1 else 1
    return c_int(c2stuff[t_interface].c_size*nele)

import array
def bytes2interface(t_interface,bytes):

  if is_string(t_interface):
    return bytes

  try:
    code = c2stuff[t_interface].py_array_code
  except KeyError:
    raise KeyError("Cannot convert %s into array"%t_interface) 
  else:  
    return array.array(code,bytes)
  
#Interface2byte is create by pyzmq

#We need the len because, if the data_interface is a c_pointer it's not iterable!
#We always receive a array. But if we now it is a scalar (len_=-1) just return
#the first element
def interface2py(data_interface,len_=-1):
  if len_ != -1:
    return [data_interface[i] for i in range(len_)]
  else:
    return data_interface[0]

#Convert:
# string    -> string
# py_scalar -> Array
# py_list   -> Array
def py2interface(t_interface,data_py):

  if is_string(t_interface):
    return data_py

  try:
    code = c2stuff[t_interface].py_array_code
  except KeyError:
    raise KeyError("Cannot convert %s into array"%t_interface) 
    
  if type(data_py) is list:  
    data_py_array = data_py
  else:
    data_py_array = [data_py]

  return array.array(code,data_py_array)  

def py2ascii(data_py):
  try:
    l = map(str,data_py)
  except TypeError:
    return str(data_py)
  else:
    return ",".join(l)

def ascii2py(t_interface,data_ascii,len_=-1):

  if is_raw(t_interface):
    return data_ascii

  try:
    code = c2stuff[t_interface].str2py
  except KeyError:
    raise KeyError("Cannot convert %s into py"%t_interface)

  if is_char(t_interface):
    #We do padding
    return code(data_ascii, t_interface)
  else:
    l_data = data_ascii.split(',')
    return map(code,l_data)

def ascii2interface(t_interface,data_ascii):

  data_py = ascii2py(t_interface,data_ascii)
  return py2interface(t_interface,data_py)

def type_fortran2c(t_fortran):
    for ctype, t in c2stuff.iteritems():
        if t_fortran.lower() == t.fortran_type.lower():
            return ctype
    raise AttributeError, "No C type for Fortran type: %s" % t_fortran

